@@ -1,4 +1,5 @@
 # -*- coding: utf-8 -*-
+import inspect
 from typing import TYPE_CHECKING
 
 from IPython import get_ipython
@@ -16,12 +17,13 @@ class Scope(object):
     def __init__(
             self, scope_name: str = GLOBAL_SCOPE_NAME,
             parent_scope: 'Optional[Scope]' = None,
-            is_namespace_scope=False
+            is_namespace_scope=False,
     ):
         self.scope_name = scope_name
         self.parent_scope = parent_scope  # None iff this is the global scope
+        self.cloned_from = None
         self.is_namespace_scope = is_namespace_scope
-        self.data_cell_by_name: Dict[str, DataCell] = {}
+        self._data_cell_by_name: Dict[str, DataCell] = {}
 
     def __hash__(self):
         return hash(self.full_path)
@@ -32,9 +34,8 @@ class Scope(object):
     def clone(self):
         cloned = Scope()
         cloned.__dict__ = dict(self.__dict__)
-        # we don't want copies of the data cells but aliases instead,
-        # but we still want separate dictionaries for newly created DataCells
-        cloned.data_cell_by_name = dict(self.data_cell_by_name)
+        cloned.cloned_from = self
+        cloned._data_cell_by_name = {}
         return cloned
 
     @property
@@ -50,8 +51,25 @@ class Scope(object):
     def make_child_scope(self, scope_name, is_namespace_scope=False):
         return self.__class__(scope_name, parent_scope=self, is_namespace_scope=is_namespace_scope)
 
+    def put(self, name: str, val: DataCell):
+        self._data_cell_by_name[name] = val
+
+    def lookup_data_cell_by_name_this_indentation(self, name):
+        ret = self._data_cell_by_name.get(name, None)
+        if ret is None and self.cloned_from is not None:
+            ret = self.cloned_from.lookup_data_cell_by_name_this_indentation(name)
+        return ret
+
+    def all_data_cells_this_indentation(self):
+        if self.cloned_from is None:
+            ret = {}
+        else:
+            ret = self.cloned_from.all_data_cells_this_indentation()
+        ret.update(self._data_cell_by_name)
+        return ret
+
     def lookup_data_cell_by_name(self, name):
-        ret = self.data_cell_by_name.get(name, None)
+        ret = self.lookup_data_cell_by_name_this_indentation(name)
         if ret is None and self.non_namespace_parent_scope is not None:
             ret = self.non_namespace_parent_scope.lookup_data_cell_by_name(name)
         return ret
@@ -62,7 +80,7 @@ class Scope(object):
         for name in chain.symbols:
             if isinstance(name, CallPoint):
                 break
-            dc = cur_scope.data_cell_by_name.get(name, None)
+            dc = cur_scope.lookup_data_cell_by_name_this_indentation(name)
             if dc is not None:
                 yield dc
             obj = name_to_obj.get(name, None)
@@ -71,53 +89,68 @@ class Scope(object):
             cur_scope = namespaces.get(id(obj), None)
             if cur_scope is None:
                 break
-            name_to_obj = obj.__dict__
+            name_to_obj = dict(inspect.getmembers(obj))
 
     def _upsert_and_mark_children_if_different_data_cell_type(
             self, dc: 'Union[ClassDataCell, FunctionDataCell]', name: str, deps: 'Set[DataCell]'
-    ):
-        if self.is_globally_accessible and name in self.data_cell_by_name:
-            old = self.data_cell_by_name[name]
-            # don't mark children as having stale dep unless old dep was of same type
-            old.update_deps(set(), add=False, mark_children=isinstance(old, type(dc)))
+    ) -> 'Tuple[DataCell, DataCell, Optional[int]]':
+        old_id = None
+        old_dc = None
+        if self.is_globally_accessible:
+            old_dc = self.lookup_data_cell_by_name_this_indentation(name)
+            if old_dc is not None:
+                old_id = old_dc.obj_id
+                # don't mark children as having stale dep unless old dep was of same type
+                old_dc.update_deps(set(), add=False, mark_children=isinstance(old_dc, type(dc)))
         dc.update_deps(deps, add=False)
-        self.data_cell_by_name[name] = dc
-        return dc
+        self.put(name, dc)
+        return dc, old_dc, old_id
 
-    def _upsert_function_data_cell_for_name(self, name: str, deps: 'Set[DataCell]'):
-        dc = FunctionDataCell(self.make_child_scope(name), name)
+    def _upsert_function_data_cell_for_name(self, name: str, obj_id: int, deps: 'Set[DataCell]'):
+        dc = FunctionDataCell(self.make_child_scope(name), name, obj_id)
         return self._upsert_and_mark_children_if_different_data_cell_type(dc, name, deps)
 
-    def _upsert_class_data_cell_for_name(self, name: str, deps: 'Set[DataCell]', class_scope: 'Scope'):
-        dc = ClassDataCell(class_scope, name)
+    def _upsert_class_data_cell_for_name(self, name: str, obj_id: int, deps: 'Set[DataCell]', class_scope: 'Scope'):
+        dc = ClassDataCell(class_scope, name, obj_id)
         return self._upsert_and_mark_children_if_different_data_cell_type(dc, name, deps)
 
     def upsert_data_cell_for_name(
             self,
             name: str,
+            obj_id: int,
             deps: 'Set[DataCell]',
             add=False,
             is_function_def=False,
             class_scope: 'Optional[Scope]' = None,
-    ):
+    ) -> 'Tuple[DataCell, DataCell, Optional[int]]':
         assert not (class_scope is not None and is_function_def)
         if is_function_def:
             assert not add
-            return self._upsert_function_data_cell_for_name(name, deps)
+            return self._upsert_function_data_cell_for_name(name, obj_id, deps)
         if class_scope is not None:
             assert not add
-            return self._upsert_class_data_cell_for_name(name, deps, class_scope)
-        if self.is_globally_accessible and name in self.data_cell_by_name:
-            # TODO: handle case where new dc is of different type
-            dc = self.data_cell_by_name[name]
-            dc.update_deps(deps, add=add)
-            # TODO: garbage collect old names
-            return dc
-        dc = DataCell(name, deps)
-        self.data_cell_by_name[name] = dc
+            return self._upsert_class_data_cell_for_name(name, obj_id, deps, class_scope)
+        old_id = None
+        old_dc = None
+        if self.is_globally_accessible:
+            old_dc = self.lookup_data_cell_by_name_this_indentation(name)
+            if old_dc is not None:
+                old_id = old_dc.obj_id
+                # TODO: garbage collect old names
+                # TODO: handle case where new dc is of different type
+                if name in self._data_cell_by_name:
+                    old_dc.update_deps(deps, add=add)
+                    old_dc.obj_id = obj_id
+                    return old_dc, old_dc, old_id
+                else:
+                    # in this case, we are copying from a class and should add the dc from which we are copying
+                    # as an additional dependency
+                    deps.add(old_dc)
+        dc = DataCell(name, obj_id, deps)
+        self.put(name, dc)
         for dep in deps:
             dep.children.add(dc)
-        return dc
+        return dc, old_dc, old_id
 
     @property
     def is_global(self):
