@@ -3,54 +3,77 @@ import ast
 import builtins
 from contextlib import contextmanager
 import logging
-from typing import TYPE_CHECKING
+from typing import cast, TYPE_CHECKING
 
 from ..data_cell import DataCell
 from ..scope import Scope
 
 if TYPE_CHECKING:
-    from typing import Dict, List, Set, Tuple, Union
+    from typing import Any, Dict, List, Optional, Set, Tuple, Union
+    Mutation = Tuple[int, Tuple[str, ...]]
+    MutCand = Optional[Tuple[int, int]]
+    SavedStoreData = Tuple[Scope, Any, str]
 
 logger = logging.getLogger(__name__)
 
 
 class AttributeTracingManager(object):
-    def __init__(self, namespaces: 'Dict[int, Scope]', active_scope: 'Scope'):
+    def __init__(self, namespaces: 'Dict[int, Scope]', active_scope: 'Scope', trace_event_counter: 'List[int]'):
         self.namespaces = namespaces
         self.original_active_scope = active_scope
         self.active_scope = active_scope
-        self.start_tracer_name = '_ATTR_TRACER_START'
-        self.end_tracer_name = '_ATTR_TRACER_END'
+        self.trace_event_counter = trace_event_counter
+        self.start_tracer_name = '_NBSAFETY_ATTR_TRACER_START'
+        self.end_tracer_name = '_NBSAFETY_ATTR_TRACER_END'
+        self.arg_recorder_name = '_NBSAFETY_ARG_RECORDER'
         setattr(builtins, self.start_tracer_name, self.attribute_tracer)
         setattr(builtins, self.end_tracer_name, self.expr_tracer)
-        self.ast_transformer = AttributeTracingNodeTransformer(self.start_tracer_name, self.end_tracer_name)
+        setattr(builtins, self.arg_recorder_name, self.arg_recorder)
+        self.ast_transformer = AttributeTracingNodeTransformer(
+            self.start_tracer_name, self.end_tracer_name, self.arg_recorder_name
+        )
         self.loaded_data_cells: Set[DataCell] = set()
-        self.stored_scope_qualified_names: Set[Tuple[Scope, str]] = set()
-        self.aug_stored_scope_qualified_names: Set[Tuple[Scope, str]] = set()
-        self.stack: List[Tuple[Set[Tuple[Scope, str]], Set[Tuple[Scope, str]], Scope, Scope]] = []
+        self.saved_store_data: Set[SavedStoreData] = set()
+        self.saved_aug_store_data: Set[SavedStoreData] = set()
+        self.mutations: Set[Mutation] = set()
+        self.recorded_args: Set[str] = set()
+        self.stack: List[
+            Tuple[Set[SavedStoreData], Set[SavedStoreData], Set[Mutation], MutCand, Set[str], Scope, Scope]
+        ] = []
+        self.mutation_candidate: MutCand = None
 
     def __del__(self):
         if hasattr(builtins, self.start_tracer_name):
             delattr(builtins, self.start_tracer_name)
         if hasattr(builtins, self.end_tracer_name):
             delattr(builtins, self.end_tracer_name)
+        if hasattr(builtins, self.arg_recorder_name):
+            delattr(builtins, self.arg_recorder_name)
 
     def push_stack(self, new_scope: 'Scope'):
         self.stack.append((
-            self.stored_scope_qualified_names,
-            self.aug_stored_scope_qualified_names,
+            self.saved_store_data,
+            self.saved_aug_store_data,
+            self.mutations,
+            self.mutation_candidate,
+            self.recorded_args,
             self.active_scope,
             self.original_active_scope,
         ))
-        self.stored_scope_qualified_names = set()
-        self.aug_stored_scope_qualified_names = set()
+        self.saved_store_data = set()
+        self.saved_aug_store_data = set()
+        self.mutations = set()
+        self.recorded_args = set()
         self.original_active_scope = new_scope
         self.active_scope = new_scope
 
     def pop_stack(self):
         (
-            self.stored_scope_qualified_names,
-            self.aug_stored_scope_qualified_names,
+            self.saved_store_data,
+            self.saved_aug_store_data,
+            self.mutations,
+            self.mutation_candidate,
+            self.recorded_args,
             self.active_scope,
             self.original_active_scope,
         ) = self.stack.pop()
@@ -61,6 +84,8 @@ class AttributeTracingManager(object):
         return obj
 
     def attribute_tracer(self, obj, attr, ctx, override_active_scope):
+        if obj is None:
+            return None
         obj_id = id(obj)
         scope = self.namespaces.get(obj_id, None)
         # print('%s attr %s of obj %s' % (ctx, attr, obj))
@@ -81,33 +106,52 @@ class AttributeTracingManager(object):
         if scope is None:
             return obj
         if ctx == 'Load':
-            data_cell = scope.data_cell_by_name.get(attr, None)
+            # save off event counter and obj_id
+            # if event counter didn't change when we process the Call retval, and if the
+            # retval is None, this is a likely signal that we have a mutation
+            # TODO: this strategy won't work if the arguments themselves lead to traced function calls
+            self.mutation_candidate = (self.trace_event_counter[0], obj_id)
+            data_cell = scope.lookup_data_cell_by_name_this_indentation(attr)
             if data_cell is None:
-                data_cell = DataCell(attr)
-                scope.data_cell_by_name[attr] = data_cell
+                data_cell = DataCell(attr, id(getattr(obj, attr, None)))
+                scope.put(attr, data_cell)
             self.loaded_data_cells.add(data_cell)
         if ctx == 'Store':
-            self.stored_scope_qualified_names.add((scope, attr))
+            self.saved_store_data.add((scope, obj, attr))
         if ctx == 'AugStore':
-            self.aug_stored_scope_qualified_names.add((scope, attr))
+            self.saved_aug_store_data.add((scope, obj, attr))
         return obj
 
     def expr_tracer(self, obj):
         # print('reset active scope to', self.original_active_scope)
+        if self.mutation_candidate is not None:
+            evt_counter, obj_id = self.mutation_candidate
+            self.mutation_candidate = None
+            if evt_counter == self.trace_event_counter[0] and obj is None:
+                self.mutations.add((obj_id, tuple(self.recorded_args)))
         self.active_scope = self.original_active_scope
+        self.recorded_args = set()
+        return obj
+
+    def arg_recorder(self, obj, name):
+        self.recorded_args.add(name)
         return obj
 
     def reset(self):
         self.loaded_data_cells = set()
-        self.stored_scope_qualified_names = set()
-        self.aug_stored_scope_qualified_names = set()
+        self.saved_store_data = set()
+        self.saved_aug_store_data = set()
+        self.mutations = set()
+        self.mutation_candidate = None
         self.active_scope = self.original_active_scope
 
 
+# TODO: handle subscripts
 class AttributeTracingNodeTransformer(ast.NodeTransformer):
-    def __init__(self, start_tracer: str, end_tracer: str):
+    def __init__(self, start_tracer: str, end_tracer: str, arg_recorder: str):
         self.start_tracer = start_tracer
         self.end_tracer = end_tracer
+        self.arg_recorder = arg_recorder
         self.inside_attr_load_chain = False
 
     @contextmanager
@@ -147,6 +191,18 @@ class AttributeTracingNodeTransformer(ast.NodeTransformer):
         assert isinstance(node.func.ctx, ast.Load)
         with self.attribute_load_context():
             node.func = self.visit_Attribute(node.func)
+        replacement_args = []
+        for arg in node.args:
+            if isinstance(arg, ast.Name):
+                replacement_args.append(cast(ast.expr, ast.Call(
+                    func=ast.Name(self.arg_recorder, ctx=ast.Load()),
+                    args=[arg, ast.Str(arg.id)],
+                    keywords=[]
+                )))
+                ast.copy_location(replacement_args[-1], arg)
+            else:
+                replacement_args.append(arg)
+        node.args = replacement_args
         if self.inside_attr_load_chain:
             return node
         replacement_node = ast.Call(
