@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 import ast
 from contextlib import contextmanager
+import logging
 from typing import TYPE_CHECKING
 
 from ..analysis import get_statement_lval_and_rval_symbols
@@ -13,6 +14,8 @@ if TYPE_CHECKING:
     from ..safety import DependencySafety
     from ..scope import Scope
 
+logger = logging.getLogger(__name__)
+
 
 class TraceStatement(object):
     def __init__(self, safety: 'DependencySafety', frame: 'FrameType', stmt_node: 'ast.stmt', scope: 'Scope'):
@@ -33,7 +36,7 @@ class TraceStatement(object):
 
     def compute_rval_dependencies(self, rval_symbols=None):
         if rval_symbols is None:
-            _, rval_symbols = get_statement_lval_and_rval_symbols(self.stmt_node)
+            _, rval_symbols, _ = get_statement_lval_and_rval_symbols(self.stmt_node)
         rval_data_cells = set()
         for name in rval_symbols:
             maybe_rval_dc = self.scope.lookup_data_cell_by_name(name)
@@ -61,18 +64,42 @@ class TraceStatement(object):
             raise TypeError('got non-function data cell %s for name %s' % (func_cell, func_name))
         return func_cell.scope
 
+    def _get_obj_id_for_name(self, name):
+        try:
+            return id(self.frame.f_locals[name])
+        except KeyError:
+            logger.error('unable to find object for name %s', name)
+            return id(None)
+
+    def _handle_aliases(
+            self,
+            old_id: 'Optional[int]', old_dc: 'Optional[DataCell]',
+            obj_id: 'Optional[int]', dc: 'Optional[DataCell]'
+    ):
+        if old_id is not None and old_dc is not None:
+            self.safety.aliases[old_id].discard(old_dc)
+        if obj_id is not None and dc is not None:
+            self.safety.aliases[obj_id].add(dc)
+        if old_id == obj_id:
+            for alias_dc in self.safety.aliases[obj_id]:
+                alias_dc.update_deps(set(), add=True)
+
     def make_lhs_data_cells_if_has_lval(self):
-        if not self.has_lval:
-            assert len(self.safety.attr_trace_manager.stored_scope_qualified_names) == 0
-            assert len(self.safety.attr_trace_manager.aug_stored_scope_qualified_names) == 0
-            return
         if not self.safety.dependency_tracking_enabled:
             return
-        lval_symbols, rval_symbols = get_statement_lval_and_rval_symbols(self.stmt_node)
+        # TODO: move handling of mutations somewhere else (doesn't fit into lval paradigm)
+        for mutated_obj_id, mutation_args in self.safety.attr_trace_manager.mutations:
+            mutation_arg_dcs = set(self.scope.lookup_data_cell_by_name(arg) for arg in mutation_args) - {None}
+            for mutated_dc in self.safety.aliases[mutated_obj_id]:
+                mutated_dc.update_deps(mutation_arg_dcs, add=True)
+        if not self.has_lval:
+            assert len(self.safety.attr_trace_manager.saved_store_data) == 0
+            assert len(self.safety.attr_trace_manager.saved_aug_store_data) == 0
+            return
+        lval_symbols, rval_symbols, should_add = get_statement_lval_and_rval_symbols(self.stmt_node)
         rval_deps = self.compute_rval_dependencies(rval_symbols=rval_symbols - lval_symbols)
         is_function_def = isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef))
         is_class_def = isinstance(self.stmt_node, ast.ClassDef)
-        should_add = isinstance(self.stmt_node, ast.AugAssign)
         if is_function_def or is_class_def:
             assert len(lval_symbols) == 1
             assert not lval_symbols.issubset(rval_symbols)
@@ -84,17 +111,28 @@ class TraceStatement(object):
                 self.safety.namespaces[id(class_ref)] = self.class_scope
             # if is_function_def:
             #     print('create function', name, 'in scope', self.scope)
-            self.scope.upsert_data_cell_for_name(
-                name, rval_deps, add=should_add_for_name, is_function_def=is_function_def, class_scope=self.class_scope
+            obj_id = self._get_obj_id_for_name(name)
+            dc, old_dc, old_id = self.scope.upsert_data_cell_for_name(
+                name, obj_id, rval_deps,
+                add=should_add_for_name, is_function_def=is_function_def, class_scope=self.class_scope
             )
-        if len(self.safety.attr_trace_manager.stored_scope_qualified_names) > 0:
+            self._handle_aliases(old_id, old_dc, obj_id, dc)
+        if len(self.safety.attr_trace_manager.saved_store_data) > 0:
             assert isinstance(self.stmt_node, ast.Assign)
-        if len(self.safety.attr_trace_manager.aug_stored_scope_qualified_names) > 0:
+        if len(self.safety.attr_trace_manager.saved_aug_store_data) > 0:
             assert isinstance(self.stmt_node, ast.AugAssign)
-        for scope, name in self.safety.attr_trace_manager.stored_scope_qualified_names:
-            scope.upsert_data_cell_for_name(name, rval_deps, add=False, is_function_def=False, class_scope=None)
-        for scope, name in self.safety.attr_trace_manager.aug_stored_scope_qualified_names:
-            scope.upsert_data_cell_for_name(name, rval_deps, add=True, is_function_def=False, class_scope=None)
+        for scope, obj, attr in self.safety.attr_trace_manager.saved_store_data:
+            obj_id = id(getattr(obj, attr, None))
+            dc, old_dc, old_id = scope.upsert_data_cell_for_name(
+                attr, obj_id, rval_deps, add=False, is_function_def=False, class_scope=None
+            )
+            self._handle_aliases(old_id, old_dc, obj_id, dc)
+        for scope, obj, attr in self.safety.attr_trace_manager.saved_aug_store_data:
+            obj_id = id(getattr(obj, attr, None))
+            dc, old_dc, old_id = scope.upsert_data_cell_for_name(
+                attr, obj_id, rval_deps, add=True, is_function_def=False, class_scope=None
+            )
+            self._handle_aliases(old_id, old_dc, obj_id, dc)
 
     def finished_execution_hook(self):
         if self.marked_finished:
