@@ -349,7 +349,6 @@ def foo(y=x):
     assert_detected("Should have detected stale dependency of fn foo() on x")
 
 
-@skipif_known_failing
 def test_same_pointer():
     # a and b are actually pointing to the same thing
     run_cell('a = [7]')
@@ -358,9 +357,9 @@ def test_same_pointer():
 
     run_cell('a[0] = 8')
     run_cell('logging.info(b)')
-    assert_not_detected("b is an alias of a, updating a should automatically update b as well")
+    assert_not_detected('`b` is an alias of `a`, updating a should automatically update `b` as well')
     run_cell('logging.info(c)')
-    assert_detected("c does not point to the same thing as a or b, thus there is a stale dependency here ")
+    assert_detected('`c` does not point to the same thing as `a` or `b`, thus there is a stale dependency here')
 
 
 def test_func_assign_objs():
@@ -623,6 +622,9 @@ class Foo(object):
             
         def foo(self):
             return Foo(self.a, self.b)
+            
+        def bar(self):
+            return Foo
 """)
     run_cell('foo = Foo(5, 6)')
     run_cell('bar = Foo.Bar(7, 8)')
@@ -631,6 +633,9 @@ class Foo(object):
     assert_not_detected()
     run_cell('Foo.Bar(9, 10).foo().shared = 100')
     run_cell('logging.info(foo.y)')
+    assert_not_detected('we mutated `shared` on obj and not on `Foo` (the one on which `foo.y` depends)')
+    run_cell('Foo.Bar(9, 10).bar().shared = 100')
+    run_cell('logging.info(foo.y)')
     assert_detected('we mutated a shared value on which `foo.y` depends')
 
 
@@ -652,7 +657,20 @@ def test_subscript_sensitivity():
     assert_detected('lst depends on stale i')
 
 
-@skipif_known_failing
+def test_subscript_adds():
+    run_cell('x = 42')
+    run_cell("""
+d = {
+    'foo': x,
+    'bar': 77
+}
+""")
+    run_cell('d["bar"] = 99')
+    run_cell('x = 100')
+    run_cell('logging.info(d)')
+    assert_detected('`d` depends on stale `x`')
+
+
 def test_list_mutation():
     run_cell('lst = list(range(5))')
     run_cell('x = 42')
@@ -662,6 +680,33 @@ def test_list_mutation():
     assert_detected('lst depends on stale x')
 
 
+def test_list_mutation_2():
+    run_cell('lst = list(range(5))')
+    run_cell('x = lst + [42, 43]')
+    run_cell('lst.append(99)')
+    run_cell('logging.info(x)')
+    assert_detected('`x` depends on stale `lst`')
+
+
+def test_lazy_class_scope_resolution():
+    run_cell("""
+class Foo(object):
+    shared = 99
+    def __init__(self, x):
+        self.x = x
+""")
+    run_cell('foo = Foo(10)')
+    run_cell('y = 11')
+    run_cell('Foo.shared = y + 42')
+    run_cell('y = 12')
+    run_cell('logging.info(foo.shared)')
+    assert_detected('`foo.shared` should point to same DataCell as `Foo.shared` and thus also has stale dep')
+    run_cell('foo.shared = 89')
+    run_cell('logging.info(Foo.shared)')
+    assert_detected('Even though we refreshed `foo.shared`, this '
+                    'has no bearing on the original class member `Foo.shared`')
+
+
 def test_numpy_subscripting_fp():
     run_cell('import numpy as np')
     run_cell('x = np.zeros(5)')
